SELECT
     PROCEDURE_CATALOG  AS DATABASE_NAME
    ,PROCEDURE_SCHEMA   AS SCHEMA_NAME
    ,PROCEDURE_NAME     AS OBJECT_NAME
    ,ARGUMENT_SIGNATURE AS SIGNATURE
    ,DATA_TYPE
    ,CHARACTER_MAXIMUM_LENGTH
    ,CHARACTER_OCTET_LENGTH
    ,NUMERIC_PRECISION
    ,NUMERIC_PRECISION_RADIX
    ,NUMERIC_SCALE
    ,PROCEDURE_LANGUAGE AS LANGUAGE
    ,PROCEDURE_DEFINITION AS DEFINITION
    ,COMMENT
FROM
    "INFORMATION_SCHEMA"."PROCEDURES"
WHERE SCHEMA_NAME = '{schema_name}';
