SELECT
     SEQUENCE_CATALOG   AS DATABASE_NAME
    ,SEQUENCE_SCHEMA    AS SCHEMA_NAME
    ,SEQUENCE_NAME      AS OBJECT_NAME
    ,DATA_TYPE
    ,NUMERIC_PRECISION
    ,NUMERIC_PRECISION_RADIX
    ,NUMERIC_SCALE
    ,START_VALUE
    ,"INCREMENT"
    ,COMMENT
FROM
    "INFORMATION_SCHEMA"."SEQUENCES"
WHERE SCHEMA_NAME = '{schema_name}';
