SELECT
     FUNCTION_CATALOG   AS DATABASE_NAME
    ,FUNCTION_SCHEMA    AS SCHEMA_NAME
    ,FUNCTION_NAME      AS OBJECT_NAME
    ,ARGUMENT_SIGNATURE AS SIGNATURE
    ,DATA_TYPE
    ,CHARACTER_MAXIMUM_LENGTH
    ,CHARACTER_OCTET_LENGTH
    ,NUMERIC_PRECISION
    ,NUMERIC_PRECISION_RADIX
    ,NUMERIC_SCALE
    ,FUNCTION_LANGUAGE  AS LANGUAGE
    ,FUNCTION_DEFINITION AS DEFINITION
    ,IS_EXTERNAL
    ,IS_SECURE
    ,VOLATILITY
    ,IS_NULL_CALL
    ,COMMENT
    ,IS_EXTERNAL
    ,API_INTEGRATION
    ,CONTEXT_HEADERS
    ,MAX_BATCH_ROWS
    ,COMPRESSION
FROM
    "INFORMATION_SCHEMA"."FUNCTIONS"
WHERE SCHEMA_NAME = '{schema_name}';
