SELECT
     PIPE_CATALOG AS DATABASE_NAME
    ,PIPE_SCHEMA  AS SCHEMA_NAME
    ,PIPE_NAME    AS OBJECT_NAME
    ,DEFINITION
    ,IS_AUTOINGEST_ENABLED
    ,NOTIFICATION_CHANNEL_NAME
    ,COMMENT
    ,PATTERN
FROM
    "INFORMATION_SCHEMA"."PIPES"
WHERE SCHEMA_NAME = '{schema_name}';
